from autochat.model import Message, MessagePart
from abc import ABC, abstractmethod
from enum import Enum
import typing


class APIProvider(Enum):
    OPENAI = "openai"
    ANTHROPIC = "anthropic"
    OPENAI_HACK = "openai_hack"


class BaseProvider(ABC):
    def prepare_messages(
        self,
        transform_function: typing.Callable,
        transform_list_function: typing.Callable = lambda x: x,
    ) -> list[dict]:
        """Prepare messages for API requests using a transformation function."""
        first_message = self.chat.messages[0]
        if self.chat.context:
            # Add context to the first message
            if isinstance(first_message.content, str):
                first_message.parts[0].content = (
                    self.chat.context + "\n" + first_message.parts[0].content
                )
            elif isinstance(first_message.content, list):
                first_message.content = [
                    MessagePart(type="text", content=self.chat.context),
                    *first_message.content,
                ]
        messages = self.chat.examples + [first_message] + self.chat.messages[1:]
        transform_list_function(messages)
        return [transform_function(m) for m in messages]

    @abstractmethod
    def fetch(self, **kwargs) -> Message:
        """
        Given a chat context, returns a single new Message from the LLM.
        """
        pass



import pytest
from PIL import Image

from autochat import Autochat, Message, APIProvider


def report_mileage(km: int):
    """Report the mileage of the car (in km)"""
    print(f"The car has {km} km on the odometer.")
    return f"The car has {km} km on the odometer."


@pytest.mark.vcr
def test_openai_read_image():
    chat = Autochat(provider=APIProvider.OPENAI)
    chat.add_function(report_mileage)

    image = Image.open("tests/images/mileage.jpg")
    message = Message(
        role="user",
        content="report the  mileage of the car",
        image=image,
    )
    response = chat.ask(
        message,
        # TODO: support tools in openai to be able to enforce the tool call
        # tool_choice={"type": "function", "function": {"name": "report_mileage"}},
    )
    print(response.to_markdown())
    # assert response.function_call["name"] == "report_mileage"
    # assert isinstance(response.function_call["arguments"].get("km"), int)


@pytest.mark.vcr
def test_anthropic_read_image():
    chat = Autochat(provider=APIProvider.ANTHROPIC)
    chat.add_function(report_mileage)

    image = Image.open("tests/images/mileage.jpg")
    message = Message(
        role="user",
        content="report the  mileage of the car",
        image=image,
    )
    response = chat.ask(message, tool_choice={"type": "tool", "name": "report_mileage"})
    assert response.function_call["name"] == "report_mileage"
    assert isinstance(response.function_call["arguments"].get("km"), int)



import anthropic
from autochat.providers.base_provider import BaseProvider
from autochat.model import Message, MessagePart
from autochat.base import AutochatBase


def part_to_anthropic_dict(part: MessagePart) -> dict:
    if part.type == "text":
        return {
            "type": "text",
            "text": part.content,
        }
    elif part.type == "image":
        return {
            "type": "image",
            "source": {
                "type": "base64",
                "media_type": part.image.format,
                "data": part.image.to_base64(),
            },
        }
    elif part.type == "function_call":
        return {
            "type": "tool_use",
            "id": part.function_call_id,
            "name": part.function_call["name"],
            "input": part.function_call["arguments"],
        }
    elif part.type == "function_result":
        return {
            "type": "tool_result",
            "tool_use_id": part.function_call_id,
            "content": part.content,
        }
    return {}


def message_to_anthropic_dict(message: Message) -> dict:
    res = {
        "role": message.role if message.role in ["user", "assistant"] else "user",
        "content": [],
    }

    for part in message.parts:
        res["content"].append(part_to_anthropic_dict(part))

    return res


class AnthropicProvider(BaseProvider):
    def __init__(self, chat: AutochatBase, model: str):
        self.model = model
        self.client = anthropic.Anthropic(
            default_headers={"anthropic-beta": "prompt-caching-2024-07-31"}
        )
        self.chat = chat

    def fetch(self, **kwargs):
        def add_empty_function_result(messages):
            """
            Ajustement Anthropic pour fusionner ou insérer la « function_result » :

            - Premier cas : message `role="function"` suivi d’un `role="user"`.
            On transforme ce message 'function' en un part de type 'tool_result'
            inséré au début du message utilisateur suivant.

            - Second cas (inchangé) : message `role="assistant"` avec un `function_call`,
            suivi d’un message non-`function`. On insère un message vide `role="function"`.
            """
            for i in range(len(messages) - 1, 0, -1):
                if (
                    messages[i - 1].role == "assistant"
                    and messages[i - 1].function_call
                    and not messages[i].role == "function"
                ):
                    # Insert an empty function result
                    messages.insert(
                        i,
                        Message(
                            role="function",
                            name=messages[i - 1].function_call["name"],
                            content="",
                            function_call_id=messages[i - 1].function_call_id,
                        ),
                    )

        messages = self.prepare_messages(
            transform_function=lambda m: message_to_anthropic_dict(m),
            transform_list_function=add_empty_function_result,
        )

        if self.chat.instruction:
            system = self.chat.instruction
        else:
            system = None

        def merge_messages(messages):
            """
            When two messages are in the same role, we merge the following message into the previous.
            {
                "role": "user",
                "content": [
                    {
                    "type": "tool_result",
                    "tool_use_id": "example_19",
                    "content": ""
                    }
                ]
            },
            {
                "role": "user",
                "content": "Plot distribution of stations per city"
            }
            """
            merged_messages = []
            for message in messages:
                if merged_messages and merged_messages[-1]["role"] == message["role"]:
                    if isinstance(merged_messages[-1]["content"], str):
                        merged_messages[-1]["content"].append(
                            {
                                "type": "text",
                                "text": merged_messages[-1]["content"],
                            }
                        )
                    elif isinstance(merged_messages[-1]["content"], list):
                        merged_messages[-1]["content"].extend(message["content"])
                    else:
                        raise ValueError(
                            f"Invalid content type: {type(merged_messages[-1]['content'])}"
                        )
                else:
                    merged_messages.append(message)
            return merged_messages

        messages = merge_messages(messages)

        # Need to map field "parameters" to "input_schema"
        tools = [
            {
                "name": s["name"],
                "description": s["description"],
                "input_schema": s["parameters"],
            }
            for s in self.chat.functions_schema
        ]
        # Add description to the function is their description is empty
        for tool in tools:
            if not tool["description"]:
                tool["description"] = "No description provided"

        # === Add cache_controls ===
        # Messages: Find the last message with an index multiple of 10
        last_message_index = next(
            (i for i in reversed(range(len(messages))) if i % 10 == 0),
            None,
        )

        if last_message_index is not None:
            if isinstance(messages[last_message_index]["content"], list) and isinstance(
                messages[last_message_index]["content"][-1], dict
            ):
                messages[last_message_index]["content"][-1]["cache_control"] = {
                    "type": "ephemeral"
                }
            elif isinstance(messages[last_message_index]["content"], str):
                messages[last_message_index]["content"] = [
                    {
                        "type": "text",
                        "text": messages[last_message_index]["content"],
                        "cache_control": {"type": "ephemeral"},
                    }
                ]
        # Tools: Add cache_control to the last tool function
        if tools:
            tools[-1]["cache_control"] = {"type": "ephemeral"}

        # System: add cache_control to the system message
        if system is not None:
            system = [
                {
                    "type": "text",
                    "text": system,
                    "cache_control": {"type": "ephemeral"},
                }
            ]

        # === End of cache_control ===

        if system is not None:
            kwargs["system"] = system

        res = self.client.messages.create(
            model=self.model,
            messages=messages,
            tools=tools,
            max_tokens=2000,
            **kwargs,
        )
        res_dict = res.to_dict()
        return Message.from_anthropic_dict(
            role=res_dict["role"],
            content=res_dict["content"],
        )



import pytest
from autochat import Autochat, APIProvider

# Our "book" data
BOOKS = {
    "The Great Gatsby": "Lorem ipsum dolor sit amet, consectetur adipiscing elit. Sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.",
    "To Kill a Mockingbird": "Lorem ipsum dolor sit amet, consectetur adipiscing elit. Sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.",
    "1984": "Lorem ipsum dolor sit amet, consectetur adipiscing elit. Sed do eiusmod tempor incididunt ut labore et dolore magna aliqua.",
}


def list_books():
    return list(BOOKS.keys())


def read_book(book_name: str):
    return BOOKS[book_name]


@pytest.mark.vcr
def test_anthropic_read_book():
    chat = Autochat(
        "You are a helpful assistant that can read books.",
        #
        provider=APIProvider.ANTHROPIC,
    )
    chat.add_function(list_books)
    chat.add_function(read_book)

    user_query = "Hello, read book 'the great gatsby'"
    messages = []
    for message in chat.run_conversation(user_query):
        print(message.to_markdown())  # For debugging
        messages.append(message)


@pytest.mark.vcr
def test_openai_read_book():
    chat = Autochat(
        "You are a helpful assistant that can read books.",
        model="gpt-4o",
        provider=APIProvider.OPENAI,
    )
    chat.add_function(list_books)
    chat.add_function(read_book)

    user_query = "Hello, read book 'the great gatsby'"
    messages = []
    for message in chat.run_conversation(user_query):
        print(message.to_markdown())  # For debugging
        messages.append(message)



"""
Temporary provider so we can use function calls with OpenAI models
that don't support them yet in the API (ie. o1).
"""

from autochat.providers.openai import OpenAIProvider, parts_to_openai_dict
from autochat.model import Message
import json
import typing


def parse_function_call_from_text(text: str):
    import re

    """
    Look for something like:

    CALLING FUNCTION: my_function
    {
        "foo": "bar"
    }

    Return (function_name, arguments_dict) or (None, None) if no function call found.
    This is a simplistic regex approach. Adjust as needed.
    """
    pattern = r"(?s)CALLING\s+FUNCTION:\s*([A-Za-z0-9_]+)\s*\n?\s*(\{.*?\})"
    match = re.search(pattern, text)
    if match:
        function_name = match.group(1).strip()
        try:
            # Attempt to parse JSON in the curly braces
            arguments = json.loads(match.group(2))
        except json.JSONDecodeError:
            arguments = {}
        return function_name, arguments
    return None, None


def from_openai_dict(
    role: str,
    content: str,
    id: typing.Optional[str] = None,
):
    function_name, function_args = parse_function_call_from_text(content)
    function_call = None
    if function_name:
        function_call = {
            "name": function_name,
            "arguments": function_args,
        }
        # Shortcut
        content = None
    return Message(role=role, content=content, function_call=function_call, id=id)


def message_to_openai_dict(message: Message) -> dict:
    """o1 only supports text-based instructions for function usage"""
    res = {
        "role": "user" if message.role == "function" else message.role,
        "content": [],
    }

    for part in message.parts:
        if part.type == "function_call":
            text = f"CALLING FUNCTION: {part.function_call['name']}\n"
            text += f"{json.dumps(part.function_call['arguments'])}\n"
            res["content"].append({"type": "text", "text": text})
        else:
            res["content"].append(parts_to_openai_dict(part))

    return res


class OpenAIProviderHack(OpenAIProvider):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def fetch(self):
        """
        Calls the OpenAI ChatCompletion endpoint using ONLY text-based instructions for function usage.
        """
        messages = self.prepare_messages(transform_function=message_to_openai_dict)
        # Add instruction as the first message
        if self.chat.instruction:
            instruction_message = Message(
                role="user",
                content=self.chat.instruction,
            )
            messages = [message_to_openai_dict(instruction_message)] + messages

        # Inject function schema into the system message, if we haven't already
        # or just append it to an existing system message. Example:
        if self.chat.functions_schema:
            function_schemas_text = json.dumps(self.chat.functions_schema, indent=2)
            augmented_instruction = (
                (self.chat.instruction or "")
                + "\n\nHere are the available functions you can call, in JSON format:\n"
                + function_schemas_text
                + "\n\n"
                + "When you call a function, respond with:\n"
                + "```\n"
                + "CALLING FUNCTION: <function_name>\n"
                + "{\n"
                + '  "arg1": "...",\n'
                + '  "arg2": "..."\n'
                + "}\n"
                + "```\n"
                + "If you do not need to call a function, just respond normally."
            )
        else:
            augmented_instruction = self.chat.instruction

        if augmented_instruction:
            system_msg = Message(
                role="user",
                content=augmented_instruction,
            )
            messages = [message_to_openai_dict(system_msg)] + messages

        # Make the request WITHOUT 'functions='
        res = self.client.chat.completions.create(
            model=self.model,
            messages=messages,
        )
        # OpenAI returns a structure. We'll parse out the text content
        message = res.choices[0].message
        # Note that now there is no function_call field because we didn't pass `functions=`.
        return from_openai_dict(
            role=message.role,
            content=message.content,
            # no function_call here
            id=res.id,
        )
